<copyright> TESLA C code.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: C.t,v 1.30 1999/10/05 19:39:55 tiggr Exp $</id>
    </copyright>

/******************** C ********************/

implementation class
C: tesla, Common
{
  <doc> Translation from byte to a character allowed in a C identifier.
      Digits are not modified.  </doc>
  static ByteString identifier_safe;

  <doc> A pointer to the elements of the above.  </doc>
  static pointer identifier_safe_chars;
}

<doc> Fill {identifier_safe}.  </doc>
void
  load Array arguments
{
  identifier_safe = "________________________________________________0123456789_______ABCDEFGHIJKLMNOPQRSTUVWXYZ______abcdefghijklmnopqrstuvwxyz______________________________________________________________________________________________________________________________________";
  (identifier_safe_chars, ) = [identifier_safe byteStringContents];
}

OutputStream (s)
  write OutputStream s
  CName String name
{
  pointer p, safe = identifier_safe_chars;
  int i, n;
  byte b;

  (p, n) = [ByteString (name) byteStringContents];
  for (i = 0; i < n; i++)
    {
      /* Naughty fast C loop.  */
      <c>
	b = ((char *) safe)[((unsigned char *) p)[i]];
      </c>
      [s print b];
    }
}

end;

implementation instance
C

<doc> Add the receiving object to the {array}.  </doc>
MutableArray (a)
  collectFlatElements MutableArray a
{
  [a add self];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[s nl self] print ("/* UNIMPLEMENTED ", self, " */")];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
      prefix String prefix
     postfix String postfix
   skipFirst boolean skip_first
{
  if (!skip_first)
    {
      if (prefix != nil)
	[s print prefix];
      [self writeCFile s];
      if (postfix != nil)
	[s print postfix];
    }
}

end;

/******************** Casgn ********************/

implementation class
Casgn: Cexpr

end;

implementation instance
Casgn
{
  Cexpr lhs, rhs;
}

id
  init (Cexpr, Cexpr) (l, r)
pre
  !!l && !!r
{
  (lhs, rhs) = (l, r);
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [s nl self];
  [lhs writeCFile s];

  // This magic should not be here.
  // Mon Jul 27 17:00:56 1998, tiggr@gerbil.org
  boolean oicvar = [lhs isRefICVar];
  if (oicvar)
    [s print (" = trt_assign_object_var (",
	      [lhs refICVarOwner], ", (void *) ")];
  else
    [s print " = "];

  [rhs writeCFile s];

  if (oicvar)
    [s print ");"];
  else
    [s print ';'];
}

end;

/******************** Cbinop ********************/

implementation class
Cbinop: Cexpr

end;

implementation instance
Cbinop
{
  Cexpr lhs, rhs;
  String op;
}

id
  init (Cexpr, String, Cexpr) (l, o, r)
pre
  !!l && !!o && !!r
{
  (lhs, op, rhs) = (l, o, r);
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [lhs writeCFile s];
  [s print (' ', op, ' ')];
  [rhs writeCFile s];
}

end;

/******************** Cblock ********************/

implementation class
Cblock: C
{
  static int temporary_count;
}

end;

implementation instance
Cblock
{
  Cblock parent;

  MutableArray decls, exprs;
}

id
  initWithParent Cblock b
{
  parent = b;
  = [super init];
}

void
  add C c
{
  if (!exprs)
    exprs = [MutableObjectArray with c];
  else
    [exprs add c];
}

void
  addDecl C c
{
  if (!decls)
    decls = [MutableObjectArray with c];
  else
    [decls add c];
}

Cvar (v)
  temporary Type t
pre
  ![t isTuple]
{
  String s = [[MutableByteString new] print ("__", temporary_count++)];
  v = [[Ctmp alloc] init (s, t)];
  [self addDecl v];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[s nl self] print '{'];
  [s indent 2];
  [self writeStartCompound s];

  int i, n;

  if (decls != nil)
    for ({i = 0; n = [decls length];}; i < n; i++)
      {
	Cvar v = decls[i];
	[[v writeCDeclaration [s nl]] print ';'];
      }

  if (exprs != nil)
    for ({i = 0; n = [exprs length];}; i < n; i++)
      [Expression (exprs[i]) writeCFile s];

  [self writeFinishCompound s];
  [s indent -2];
  [[s nl nil] print '}'];
}

TeslaOutput (s)
  writeFinishCompound TeslaOutput s
{
}

TeslaOutput (s)
  writeStartCompound TeslaOutput s
{
}

end;

/******************** CBlockDecl ********************/

implementation class
CBlockDecl: C

end;

implementation instance
CBlockDecl
{
  BlockDefinition block;
}

id (self)
  init BlockDefinition bd
{
  block = bd;
}

<doc> Declare the struct containing the block.  </doc>
TeslaOutput (of)
  writeCDeclaration TeslaOutput of
{
  [block writeBlockStructName [of print "struct "]];
  [block writeBlockName [of print " *"]];
}

end;

/******************** Cbody ********************/

implementation class
Cbody: Ccallable

end;

implementation instance
Cbody
{
  MethodBody body;
}

id
  initWithBody MethodBody b
{
  body = b;
  = [super initWithParent nil];
}

TeslaOutput (s)
  writeFinishCompound TeslaOutput s
{
  if ([body uses_return])
    [[s nl nil] print "_end_of_method:"];
  ArgumentTuple rv = [[body method] return_names];
  if (rv != nil)
    {
      [rv writeCReturnAssignments s skipFirst YES];
      Argument r = Argument ([rv tupleFirstElement]);
      if ([r type] != void_type)
	[[r writeCName [[s nl nil] print "return "]] print ';'];
    }
  temporary_count = 0;
}

end;

/******************** Ccallable ********************/

implementation class
Ccallable: Cblock

end;

implementation instance
Ccallable
{
  <doc> The extensions of which non-static state is referenced from this
      body.  </doc>
  MutableArray extensions;

  <doc> {TRUE} iff this callable reference thread-local variables.  </doc>
  boolean need_locals;
}

<doc> Note that this callable references variables from the
    extension {ext}.  </doc>
void
  needExtension Extension ext
{
  if (!extensions)
    {
      extensions = [MutableObjectArray with ext];
    }
  else if ([extensions indexOfIdentical ext] < 0)
    [extensions add ext];
}

<doc> Note that this callable references thread-local variables.
    </doc>
void
  needLocals
{
  need_locals = TRUE;
}

TeslaOutput (s)
  writeStartCompound TeslaOutput s
{
  if (extensions != nil)
    {
      int i, n = [extensions length];
      for (i = 0; i < n; i++)
	[extensions[i] writeCNameStatePointerDeclaration s
		   for current_meta];
    }

  // The thread-local pointer should not be cached...
  // Wed Jul 22 16:55:54 1998, tiggr@gerbil.org
  if (need_locals)
    [[s nl] print "char *_th_locals = trt_thread_get_local_data ();"];
}

end;

/******************** Cclosure ********************/

implementation class
Cclosure: Ccallable

end;

implementation instance
Cclosure
{
  BlockDefinition block;
}

id
  initWithBlock BlockDefinition b
{
  block = b;
  = [super initWithParent nil];
}

TeslaOutput (s)
  writeStartCompound TeslaOutput s
{
  boolean have_context = [block context] != nil;
  boolean have_vars = [block haveBlockVariables];

  if (have_context || have_vars)
    {
      [[s nl block] print ("struct _es_i_tom_Block *_block_this = ",
			   "trt_ext_address (block_self, _ei_i_tom_Block);")];
      if (have_context)
	{
	  [block writeContextStructName [[s nl block] print "struct "]];
	  [[block writeContextName [s print " *"]]
	    print " = _block_this->environment;"];
	}
      if (have_vars)
	{
	  [block writeBlockStructName [[s nl block] print "struct "]];
	  [[block writeBlockName [s print " *"]]
	    print " = _block_this->variables;"];
	}
    }

  VariableSingle vs = [block nameof_self];
  if (vs != nil)
    {
      [[s nl block] print "void *self = "];
      [[[[Ctomvar alloc vs] init vs] reference nil] writeCFile s];
      [s print ";"];
    }

  [super writeStartCompound s];
}

TeslaOutput (s)
  writeFinishCompound TeslaOutput s
{
  ArgumentTuple rv = [block eval_names];
  if (rv != nil)
    {
      [rv writeCReturnAssignments s skipFirst YES];
      Argument r = Argument ([rv tupleFirstElement]);
      if ([r type] != void_type)
	[[r writeCName [[s nl nil] print "return "]] print ';'];
    }
}

end;

/******************** Ccast ********************/

implementation class
Ccast: Cexpr

end;

implementation instance
Ccast
{
  <doc> The type.  </doc>
  Type type;

  <doc> The argument.  </doc>
  Cexpr argument;
}

id
  init (Type, Cexpr) (t, a)
{
  (type, argument) = (t, a);
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[type writeCType [s print '(']] print ") ("];
  [[argument writeCFile s] print ')'];
}

end;

/******************** Ccode ********************/

implementation class
Ccode: Cconst

end;

implementation instance
Ccode

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [super writeCFile [s nl self]];
}

end;

/******************** Cconst ********************/

implementation class
Cconst: Cexpr

end;

implementation instance
Cconst
{
  String c;
}

id 
  init String s
{
  c = s;
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [s print c];
}

end;


/******************** CContextDecl ********************/

implementation class
CContextDecl: C

end;

implementation instance
CContextDecl
{
  Callable callable;
}

id (self)
  init Callable c
{
  callable = c;
}

<doc> Declare the struct containing the context.  </doc>
TeslaOutput (of)
  writeCDeclaration TeslaOutput of
{
  MutableByteString cn = [MutableByteString new];
  [callable writeContextName cn];

  [[of print ("struct ", cn, " ", cn, "_def;")] nl];
  [of print ("struct ", cn, " *", cn, " = &", cn, "_def")];
}

end;

/******************** Cdynmi ********************/

implementation class
Cdynmi: Cmi

end;

implementation instance
Cdynmi
{
  Cexpr rcv;
}

id 
  init (Cexpr, Selector, Indexed, Cexpr, MethodInvocation) (r, s, a, l, mi)
{
  rcv = r;
  = [super init (s, a, l, mi)];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  Type t = [invocation type];
  [s nl invocation];

  /* Left hand side of the C assignment.  */
  if (t != void_type)
    [[[lhs tupleFirstElement] writeCFile s] print " = "];

  /* Cast of the result of trt_lookup.  */
  [[[t tupleFirstElement] writeCType [s print "(("]]
    print " (*) (void *, selector"];
  int i, n = !arguments ? 0 : [arguments length];
  if (n != 0)
    {
      Indexed orig_args = [invocation arguments];
      for (i = 0; i < n; i++)
	[[orig_args[i] type] writeCTypes s prefix ", " postfix nil
			     skipFirst NO];
    }
  [t writeCTypes s prefix ", " postfix "*" skipFirst YES];

  /* The lookup.  */
  [s print ")) trt_lookup ((void *) "];
  [[rcv writeCFile s] print ", &"];
  [[sel writeCReferenceName s] print ")) "];

  /* The method arguments.  */
  [sel writeCReferenceName [[rcv writeCFile [s print "("]] print ", &"]];
  for (i = 0; i < n; i++)
    [arguments[i] writeCFile [s print ", "]];
  if (lhs != nil)
    [lhs writeCFile s prefix ", &" postfix nil skipFirst YES]; 
  [s print ");"];
}

end;

/******************** Cexpr ********************/

implementation class
Cexpr: C

end;

implementation instance
Cexpr

<doc> Emit enough assignments to assign the receiving expression to the
    {lhs}.  The default implementation emits a single assignment.  </doc>
void
       compile Cblock parent
  assignmentTo Cexpr lhs
{
  [parent add [[Casgn alloc self] init (lhs, self)]];
}

<doc> Return {TRUE} if the receiving {Cexpr} is a reference to an
    instance or class variable with a reference type.  The default
    implementation returns {FALSE}.  </doc>
boolean
  isRefICVar
{
}

<doc> Compile the assignment to the receiving Cexpr of the value {rhs}.
    The default implementation simply emits a {Casgn}.  </doc>
void
    precompile Cblock parent
	method Cblock body
  assignmentOf Cexpr rhs
{
  [parent add [[Casgn alloc] init (self, rhs)]];
}

<doc> Return a string representing the owning object of a reference
    icvar.  This will either return {self} or {self->isa}.  </doc>
deferred String
  refICVarOwner;

<doc> If the receiving expression is a variable but not a temporary, emit
    an assignment of the variable to a temporary and return the temporary.
    The default implementation returns self.  </doc>
Cexpr
  temporaries Cblock parent
       method Cblock body
{
  = self;
}

Cexpr
  tupleFirstElement
{
  = self;
}

end;

/******************** Cite ********************/

implementation class
Cite: Cexpr

end;

implementation instance
Cite
{
  Cexpr cond;
  C then_block, else_block;
}

id
  init (Cexpr, C, C) (i, t, e)
{
  (cond, then_block, else_block) = (i, t, e);
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[s nl self] print "if ("];
  [[cond writeCFile s] print ')'];
  [s indent 2];
  [then_block writeCFile s];
  [s indent -2];
  if (else_block != nil)
    {
      [[s nl nil] print "else"];
      [s indent 2];
      [else_block writeCFile s];
      [s indent -2];
    }
}

end;

/******************** Cfor ********************/

implementation class
Cfor: Cexpr

end;

implementation instance
Cfor
{
  Cblock body;
}

id
  init Cblock b
{
  body = b;
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[s nl self] print "for (;;)"];
  [s indent 2];
  [body writeCFile s];
  [s indent -2];
}

end;

/******************** Clitcast ********************/

implementation class
Clitcast: Cexpr

instance (id)
  with (String, Cexpr) (t, e)
{
  = [[self alloc] init (t, e)];
}

end;

implementation instance
Clitcast
{
  <doc> The type.  </doc>
  String type;

  <doc> The argument.  </doc>
  Cexpr argument;
}

protected id
  init (String , Cexpr) (t, a)
{
  (type, argument) = (t, a);
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[argument writeCFile [s print ("(", type, ") (")]] print ')'];
}

end;

/******************** Cmi ********************/

implementation class
Cmi: Cexpr

end;

implementation instance
Cmi
{
  Cexpr lhs;
  Selector sel;
  Indexed arguments;
  MethodInvocation invocation;
}

id 
  init (Selector, Indexed, Cexpr, MethodInvocation) (s, a, l, mi)
{
  (sel, arguments, lhs, invocation) = (s, a, l, mi);
  = [super init];
}

end;

/******************** Cmkblock ********************/

implementation class
Cmkblock: Cexpr

end;

implementation instance
Cmkblock
{
  BlockDefinition block;
}

id
  initWithBlock BlockDefinition b
{
  block = b;
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [block writeBlockCodeName [s print "trt_new_block ((void (*)) "]];
  [s print ", &"];
  [[[block formal_selector] writeCReferenceName s] print ", "];
  [[Compound ([block container]) callable] writeContextNameOrNil s];

  if (![block haveBlockVariables])
    [s print ", 0, 0"];
  else
    {
      [block writeBlockName [s print ", "]];
      [block writeBlockDescriptionName [s print ", &"]];
    }

  [s print ")"];
}

end;

/******************** Cmonop ********************/

implementation class
Cmonop: Cexpr

end;

implementation instance
Cmonop
{
  Cexpr shs;
  String op;
}

id
  init (String, Cexpr) (o, s)
pre
  !!o && !!s
{
  (op, shs) = (o, s);
  = [super init];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [shs writeCFile [s print op]];
}

end;

/******************** Cref ********************/

implementation class
Cref: Cexpr

end;

implementation instance
Cref
{
  Cvar referenced;
}

id
  init Cvar r
{
  referenced = r;
  = [super init];
}

<doc> Pass on to the {referenced}.  </doc>
boolean
  isRefICVar
{
  = [referenced isRefICVar];
}

<doc> Pass on to the {referenced}.  </doc>
String
  refICVarOwner
{
  = [referenced refICVarOwner];
}

Cexpr (result)
  temporaries Cblock parent
       method Cblock body
{
  if ([referenced isTemporary])
    = self;
  else
    {
      result = [[referenced type] temporaryCexpr parent];
      [parent add [[Casgn alloc self] init (result, self)]];
    }
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [referenced writeCReference s];
}

end;

/******************** Csupmi ********************/

implementation class
Csupmi: Cmi

end;

implementation instance
Csupmi
{
  SuperReference super_ref;
}

<doc> Designated initializer.  </doc>
id 
  init (SuperReference, Selector, Indexed, Cexpr, MethodInvocation)
       (sr, s, a, l, mi)
{
  super_ref = sr;
  = [super init (s, a, l, mi)];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  Type t = [invocation type];

  [s nl invocation];

  /* Left hand side of the C assignment.  */
  if (t != void_type)
    [[[lhs tupleFirstElement] writeCFile s] print " = "];

  /* Cast of the result of trt_lookup_super.  */
  [[[t tupleFirstElement] writeCType [s print "(("]]
    print " (*) (void *, selector"];
  int i, n = !arguments ? 0 : [arguments length];
  if (n != 0)
    {
      Indexed orig_args = [invocation arguments];
      for (i = 0; i < n; i++)
	[[orig_args[i] type] writeCTypes s prefix ", " postfix nil
			     skipFirst NO];
    }
  [t writeCTypes s prefix ", " postfix "*" skipFirst YES];

  /* The lookup.  */
  [super_ref writeCFile [s print ")) trt_lookup_super ("]];
  [[sel writeCReferenceName [s print ", &"]] print ")) "];

  /* The method arguments.  */
  // SELF should not be implicit.
  // Wed Jul  8 20:29:23 1998, tiggr@gerbil.org
  [sel writeCReferenceName [s print "(self, &"]];
  for (i = 0; i < n; i++)
    [arguments[i] writeCFile [s print ", "]];
  if (lhs != nil)
    [lhs writeCFile s prefix ", &" postfix nil skipFirst YES]; 
  [s print ");"];
}

end;

/******************** Cstmt ********************/

implementation class
Cstmt: Ccode

end;

implementation instance
Cstmt

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [[super writeCFile s] print ';'];
}

end;

/******************** Ctmp ********************/

implementation class
Ctmp: Cvar

end;

implementation instance
Ctmp
{
  public Type type;
  String name;
}

id
  init (String, Type) (s, t)
{
  (name, type) = (s, t);
  = [super init];
}

<doc> Return {TRUE}.  </doc>
boolean
  isTemporary
{
  = TRUE;
}

Cref
  reference Reference r
{
  = [[Cref alloc r] init self];
}

TeslaOutput (s)
  writeCDeclaration TeslaOutput s
{
  [[type writeCType s] print (" ", name)];
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [s print name];
}

end;

/******************** Ctomvar ********************/

implementation class
Ctomvar: Cvar

end;

implementation instance
Ctomvar
{
  Variable original;
}

id
  init Variable o
{
  original = o;
  = [super init];
}

<doc> Pass to the {original}.  </doc>
boolean
  isRefICVar
{
  = [original isRefICVar];
}

<doc> Pass on to the {original}.  </doc>
String
  refICVarOwner
{
  = [original refICVarOwner];
}

Type
  type
{
  = [original type];
}

TeslaOutput (s)
  writeCDeclaration TeslaOutput s
{
  [original writeCDeclaration s];
}

TeslaOutput (s)
  writeCReference TeslaOutput s
{
  [original writeCReference s];
}

end;

/******************** Ctuple ********************/

implementation class
Ctuple: Cexpr

end;

implementation instance
Ctuple
{
  MutableArray elements;
}

id
  init int capacity
{
  elements = [MutableObjectArray withCapacity capacity];
  = [super init];
}

void
  add Cexpr e
{
  [elements add e];
}

Cexpr
  at int i
{
  = elements[i];
}

<doc> Pass to the {elements}.  </doc>
MutableArray (a)
  collectFlatElements MutableArray a
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] collectFlatElements a];
}

void
       compile Cblock parent
  assignmentTo Ctuple lhs
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [Cexpr (elements[i]) compile parent assignmentTo lhs[i]];
}

<doc> Emit assignments of the elements.  </doc>
void
    precompile Cblock parent
	method Cblock body
  assignmentOf Ctuple rhs
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] precompile parent method body assignmentOf rhs[i]];
}

<doc> Pass to the {elements} and update the receiving tuple.  </doc>
Ctuple
  temporaries Cblock parent
       method Cblock body
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    {
      Cexpr a = elements[i], b = [a temporaries parent method body];
      if (a != b)
	elements[i] = b;
    }
  = self;
}

Cexpr
  tupleFirstElement
{
  = [elements[0] tupleFirstElement];
}

redefine TeslaOutput (s)
  writeCFile TeslaOutput s
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    {
      if (i > 0)
	[s print ", "];
      [Extension (elements[i]) writeCFile s];
    }
}

redefine TeslaOutput (s)
  writeCFile TeslaOutput s
      prefix String prefix
     postfix String postfix
   skipFirst boolean skip_first
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    {
      [elements[i] writeCFile s prefix prefix postfix postfix
	       skipFirst skip_first];
      skip_first = FALSE;
    }
}

end;

/******************** Cvar ********************/

implementation class
Cvar: C

end;

implementation instance
Cvar

<doc> Return {TRUE} if the receiving variable represents an instance
    or class variable with a reference type.  The default
    implementation returns {FALSE}.  </doc>
boolean
  isRefICVar
{
}

<doc> Return {TRUE} iff the receiving variable is a temporary.  The
    default implementation returns {FALSE}.  </doc>
boolean
  isTemporary
{
}

Cref
  reference Expression r
{
  = [[Cref alloc r] init self];
}

deferred String
  refICVarOwner;

deferred Type
  type;

deferred OutputStream (s)
  writeCDeclaration OutputStream s;

<doc> Write a reference to this variable (in an expression or similar
    context).  The default implementation invokes {writeCFile}.
    </doc>
TeslaOutput (s)
  writeCReference TeslaOutput s
{
  = [self writeCFile s];
}

end;

/******************** Cxti ********************/

implementation class
Cxti: C

end;

implementation instance
Cxti
{
  ArgumentSingle arg;
}

id (self)
  init ArgumentSingle a
{
  arg = a;
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [arg writeContextInitialization s];
}

end;
