<copyright> TESLA Loop class
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998, 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Loop.t,v 1.1 1999/08/13 21:20:38 tiggr Exp $</id>
    </copyright>

/******************** Loop ********************/

implementation class
Loop: Compound
{
  <doc> Counter for labels to which continue can jump.  </doc>
  static int num_loop_labels;
}

end;

implementation instance
Loop
{
  <doc> Various control fields.  </doc>
  Expression initializer, condition, modifier;

  <doc> A temporary for the value of this loop.  </doc>
  public Cexpr cresult;

  <doc> The name of the label for continue, if set.  </doc>
  String label;
}

<doc> Designated initializer.  </doc>
id
  initWithControl (Expression, Expression, Expression) (i, c, m)
	container Context cxt
{
  (initializer, condition, modifier) = (i, c, m);
  = [super initWithContainer cxt];
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  if (initializer != nil)
    [initializer coarseUsage f & ~(EXPR_RHS | EXPR_LHS) parent c body m];
  if (condition != nil)
    [condition coarseUsage f & ~EXPR_LHS | EXPR_RHS parent c body m];
  if (modifier != nil)
    [modifier coarseUsage f & ~(EXPR_RHS | EXPR_LHS) parent c body m];
  [super coarseUsage f parent c body m];
}

<doc> Return {self}.  </doc>
Loop (self)
  enclosingLoop
{
}

<doc> Return a (newly created) {label}.  </doc>
String
  continue_label
{
  if (!label)
    label = [[MutableByteString new] print ("_label_", ++num_loop_labels)];
  = label;
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (c_flags & C_TYPED_EXPRS == 0)
    {
      if (initializer != nil)
	initializer = [initializer resolve (nil, nil) context context_type];
      if (modifier != nil)
	modifier = [modifier resolve (nil, nil) context context_type];
      if (condition != nil)
	condition = [condition resolve
		     ([Cons with (basic_types_from_kind[BT_BOOLEAN], nil)], nil)
			       context context_type];
    }

  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

Cexpr (result)
  precompile Cblock parent
      method Cblock method_body
{
  Cblock bb = [[Cblock alloc] initWithParent parent];

  if (expr_flags & EXPR_RHS != 0)
    {
      result = cresult = [type temporaryCexpr parent];
      Cexpr iv = [[type defaultValue nil] precompile parent method method_body];
      [parent add [[Casgn alloc] init (result, iv)]];
    }

  if (initializer != nil)
    [initializer precompile parent method method_body];

  Cblock cc;
  if (condition != nil)
    {
      cc = [[Cblock alloc] initWithParent bb];
      Cexpr ccc = [condition precompile cc method method_body];
      [cc add [[Cite alloc condition]
		init ([[Cmonop alloc] init ("!", ccc)],
		      [[Cstmt alloc] init "break"], nil)]];
    }

  if (cc != nil && ![self condition_after])
    [bb add cc];

  Cexpr body_result = [self precompileElements bb method method_body];
  if (result != nil)
    [bb add [[Casgn alloc body_result] init (result, body_result)]];

  if (label != nil)
    [bb add [[Ccode alloc self]
	      init [[MutableByteString new] print (label, ":")]]];

  if (modifier != nil)
    [modifier precompile bb method method_body];

  if (cc != nil && [self condition_after])
    [bb add cc];

  [parent add [[Cfor alloc] init bb]];
  cresult = nil;
}

void
  set_condition Expression c
pre
  !condition
{
  condition = c;
}

/******************** flags ********************/

boolean
  condition_after
{
  = c_flags & 1 << LOOP_CONDITION_AFTER != 0;
}

void
  set_condition_after boolean b
{
  if (b)
    c_flags |= 1 << LOOP_CONDITION_AFTER;
  else
    c_flags &= ~(1 << LOOP_CONDITION_AFTER);
}

end;
